/* ONLY FOR KERNEL 2.4 or more */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

//#define FX2000_DEBUG

//#define CONFIG_DEVFS_FS


#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/errno.h>    /* error codes */
#include <linux/version.h>
#include <linux/config.h>
#include <linux/init.h>

#include <linux/fs.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include <linux/usb.h>


#include "fx2000.h"

/*
 * Version Information
 */
#define DRIVER_VERSION "v0.02"
#define DRIVER_AUTHOR "Giorgio Alboni, alboni@biometrika.it"
#define DRIVER_DESC "Fx2000 USB interface driver for Linux (2002)"
#define BUFFER_IN_SIZE 120000
#define BUFFER_OUT_SIZE 1200

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,0)
#define KERNEL26
#else
#define KERNEL24
#endif

typedef int BOOL;
#define FALSE	0
#define TRUE	1

/* read_scanner timeouts -- RD_NAK_TIMEOUT * RD_EXPIRE = Number of seconds */
#define WR_NAK_TIMEOUT (1*HZ)	/* Default number of X seconds to wait */
#define RD_NAK_TIMEOUT (1*HZ)	/* Default number of X seconds to wait */
#define RD_EXPIRE 3		/* Number of attempts to wait X seconds */

/*
 * We need a local data structure, as it must be allocated for each new
 * mouse device plugged in the USB bus
 */
typedef struct fx2000_device {
	struct usb_device *udev;			/* save off the usb device pointer */
	struct usb_interface *interface;		/* the interface for this device */
	char bufferIn[BUFFER_IN_SIZE];
	char bufferOut[BUFFER_OUT_SIZE];
	BOOL deviceOpened;
} fx2000_device_t;

typedef fx2000_device_t *fx2000_device_p;

int f_fx2000_open(struct inode *inode, struct file *filp);
ssize_t f_fx2000_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
ssize_t f_fx2000_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);
int f_fx2000_release(struct inode *inode, struct file *filp);

static struct file_operations fx2000_fops = {
	.read=       f_fx2000_read,
	.write=      f_fx2000_write,
	.open=       f_fx2000_open,
	.release=    f_fx2000_release,
};

#ifdef CONFIG_SYSFS
static struct class_simple *sysfs_class;
#endif

fx2000_device_p dev_fx2000[FX2000_NR_DEVS];
BOOL fx2000_presence[FX2000_NR_DEVS];

int fx2000_major =   FX2000_MAJOR;
int fx2000_nr_devs = FX2000_NR_DEVS;    /* number of bare scull devices */

MODULE_PARM(fx2000_major,"i");
MODULE_PARM(fx2000_nr_devs,"i");

#ifdef FX2000_DEBUG
static void print_dev_fx2000 (void)
{
	int i;

	for (i=0; i<FX2000_NR_DEVS ;i++)
		if (dev_fx2000[i]!=NULL)
				FX2000_DEBUG("FX2000 n.%i(%4x)= %x\n",i,dev_fx2000[i]->udev->descriptor.idProduct,dev_fx2000[i]);
			else
				FX2000_DEBUG("FX2000 n.%i(%--)= %x\n",i,dev_fx2000[i]);
}
#endif

/* -------------------------------------------------------------------------------*/
/* Handle proc file																																*/
/* -------------------------------------------------------------------------------*/

/* Proc read function: print some driver information */
int fx2000_read_procmem(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
	int i, len = 0;
	//int limit = count - 80; /* Don't print more than this */

	sprintf (buf,"Fingerprint scanner Fx2000 driver running\n\nDevice status:\n");
	len=strlen (buf);

	for (i=0; i<FX2000_NR_DEVS ;i++)
	{
		if (fx2000_presence[i])
				sprintf(buf+len,"FX2000 n.%3i (%04x)= Connected (%s)\n",i?i+100:0,
								dev_fx2000[i]->udev->descriptor.idProduct,dev_fx2000[i]->deviceOpened? "open":"close");
			else
				sprintf(buf+len,"FX2000 n.%3i (----)= Not Connected\n",i?i+100:0);
		len=strlen (buf);
	}
  *eof = 1;
  return len;
}



/* Remove proc file */
static void fx2000_create_proc()
{
    create_proc_read_entry("fx2000", 0 /* default mode */,
                           NULL /* parent dir */, fx2000_read_procmem,
                           NULL /* client data */);
}

/* Register proc file */
static void fx2000_remove_proc()
{
    remove_proc_entry("fx2000", NULL /* parent dir */);
}

/* -------------------------------------------------------------------------------*/
/* Handle USB events																															*/
/* -------------------------------------------------------------------------------*/

/*
 * These two callbacks are invoked when an USB device is detached or attached
 * to the bus
 */
#ifdef KERNEL26
static void fx2000_disconnect(struct usb_interface *intf)
#else
static void fx2000_disconnect(struct usb_device *udev,void* clientdata)
#endif
{

#ifdef KERNEL26
	struct usb_device *udev=interface_to_usbdev(intf);
	struct fx2000_device *scanner = usb_get_intfdata(intf);
	usb_set_intfdata(intf,NULL);
#else
	struct fx2000_device *scanner = clientdata;	
#endif

	/* Clear reference in the fx2000 device array */
	fx2000_presence[udev->descriptor.idProduct-0x0200]=FALSE;
	dev_fx2000[udev->descriptor.idProduct-0x0200]=NULL;
	
#ifdef CONFIG_SYSFS
	class_simple_device_remove(MKDEV(fx2000_major,udev->descriptor.idProduct-0x0200));
#endif
	
	/* Free kernel memory */
	if (scanner!=NULL)
		kfree(scanner);

#ifdef FX2000_DEBUG
	FX2000_DEBUG("USB disconnected\n");
	print_dev_fx2000();
#endif

  return;
}

#ifdef KERNEL26
static int fx2000_probe(struct usb_interface *intf, 
			  const struct usb_device_id *id)
#else
static void* fx2000_probe(struct usb_device *udev,unsigned int ifnum, 
			  const struct usb_device_id *id)
#endif
{
	/*
	 * The probe procedure is pretty standard. Device matching has already
	 * been performed based on the id_table structure (defined later)
	 */
	
#ifdef KERNEL26
	struct usb_device *udev=interface_to_usbdev(intf);
#endif
	struct fx2000_device *scanner;
	

	/* We don't handle multiple configurations */
	if (udev->descriptor.bNumConfigurations != 1)
#ifdef KERNEL26
		return -ENODEV;
#else
		return NULL;
#endif

#ifdef FX2000_DEBUG
	/* Probed device info */
	FX2000_DEBUG("probe: vendor id 0x%x, device id 0x%x ifnum:%d\n",
				udev->descriptor.idVendor, udev->descriptor.idProduct, intf->minor);
#endif

	/* allocate and zero a new data structure for the new device */
	scanner = kmalloc(sizeof(struct fx2000_device), GFP_KERNEL);
	if (!scanner) 
#ifdef KERNEL26
	    return -ENOMEM; /* failure */
#else
	    return NULL; /* failure */
#endif
	memset(scanner, 0,     sizeof(*scanner));

	/* Save data to new structure */
	scanner->udev=udev;
	scanner->deviceOpened=FALSE;
	dev_fx2000[udev->descriptor.idProduct-0x0200]=scanner;
	fx2000_presence[udev->descriptor.idProduct-0x0200]=TRUE;
#ifdef CONFIG_SYSFS
	class_simple_device_add(sysfs_class,MKDEV(fx2000_major,udev->descriptor.idProduct-0x0200),NULL,"fx2000_%d",udev->descriptor.idProduct-0x0200);
#endif
	
#ifdef FX2000_DEBUG
	/* announce yourself */
	FX2000_DEBUG("PROBE successful\n");
	print_dev_fx2000();
#endif

#ifdef KERNEL26
	usb_set_intfdata(intf,scanner);
	return 0;
#else
	return scanner;
#endif
}

/* -------------------------------------------------------------------------------*/
/* USB file function																															*/
/* -------------------------------------------------------------------------------*/

int f_fx2000_open(struct inode *inode, struct file *filp)
{
	int minor=MINOR(inode->i_rdev);

#ifdef FX2000_DEBUG
	FX2000_DEBUG ("Open device (%i)",minor);
#endif
	
	/* Check device presence */
	if ((minor<0) || (minor>FX2000_NR_DEVS))
		return -ENODEV; // Wrong Minor (device number)
	if (!fx2000_presence[minor])
		return -ENODEV; // Device not probed (not present)
	if (dev_fx2000[minor]->deviceOpened==TRUE)
		return -ENODEV; // Device olready open!!! 
	/* Config data */
	filp->private_data=&fx2000_presence[minor];
	dev_fx2000[minor]->deviceOpened=TRUE;
#ifdef FX2000_DEBUG
	FX2000_DEBUG (" - OK\n");
#endif
	return 0;          /* success */
}

int f_fx2000_release(struct inode *inode, struct file *filp)
{
	int minor=MINOR(inode->i_rdev);

#ifdef FX2000_DEBUG
	FX2000_DEBUG ("Close device (%i)\n",minor);
#endif

	/* Check device presence */
	if (!fx2000_presence[minor])
		return 0;	// Something WRONG! The device was disconnect before close

	/* Remove open flag (device now available) */
	dev_fx2000[minor]->deviceOpened=FALSE;
	filp->private_data=NULL;

	return 0;
}

ssize_t f_fx2000_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
	struct usb_device *udev;
	struct fx2000_device *scanner;
	BOOL minor_presence;
	int minor;

	wait_queue_head_t wait_q;
	ssize_t bytes_read;	/* Overall count of bytes_read */
	ssize_t ret;
	int partial;		/* Number of bytes successfully read */
	int this_read;		/* Max number of bytes to read */
	int result;
	int rd_expire = RD_EXPIRE;
	int i;

	minor_presence=(BOOL)*((BOOL*)filp->private_data);

	if (!minor_presence)
		return -EFAULT;	// Something wrong: device not more available!

	// Search minor number (for dev_fx2000 index)
	for (i=0; (i<FX2000_NR_DEVS) && (&fx2000_presence[i]!=filp->private_data); i++);
	if (&fx2000_presence[i]!=filp->private_data) return -EFAULT;	// Not Find!!
	minor=i;

	scanner=dev_fx2000[minor];
	udev=scanner->udev;

	bytes_read = 0;
	ret = 0;

	filp->f_dentry->d_inode->i_atime = CURRENT_TIME;
	init_waitqueue_head(&wait_q);

	while (count > 0) {
		if (signal_pending(current)) 
		{
			ret = -EINTR;
			break;
		}
		this_read = (count >= BUFFER_IN_SIZE) ? BUFFER_IN_SIZE : count;

		result = usb_bulk_msg(udev, usb_rcvbulkpipe(udev, 5), scanner->bufferIn, this_read, &partial,(int) RD_NAK_TIMEOUT);

#ifdef FX2000_DEBUG
		FX2000_DEBUG("read stats(%d): result:%d this_read:%d partial:%d count:%d\n", minor, result, this_read, partial, count);
#endif

		if (result == -ETIMEDOUT) /* NAK */
			{
				if (!partial) /* No data */
				{
					if (--rd_expire <= 0) /* Give it up */
						{
#ifdef FX2000_DEBUG
							FX2000_DEBUG("read_scanner(%d): excessive NAK's received", minor);
#endif
							ret = result;
							break;
						} 
					else /* Keep trying to read data */
						{
							//interruptible_sleep_on_timeout(&wait_q, RD_NAK_TIMEOUT);
							continue;
						}
				} 
				else /* Timeout w/ some data */
				{
					goto data_recvd;
				}
			}		
		if (result == -EPIPE) /* No hope */
			{
				if(usb_clear_halt(udev, 5)) 
				{
#ifdef FX2000_DEBUG
					FX2000_DEBUG("read_scanner(%d): Failure to clear endpoint halt condition (%Zd).", minor, ret);
#endif
				}
				ret = result;
				break;
			} 
		else 
			if ((result < 0) && (result != -ENOSR)) 
			{
#ifdef FX2000_DEBUG
				FX2000_DEBUG("read_scanner(%d): funky result:%d. Please notify the maintainer.", minor, (int)result);
#endif
				ret = -EIO;
				break;
			}

		data_recvd:
		if (partial) /* Data returned */
			{
				if (copy_to_user(buf, scanner->bufferIn, partial)) 
					{
						ret = -EFAULT;
						break;
					}
				count -= this_read; /* Compensate for short reads */
				bytes_read += partial; /* Keep tally of what actually was read */
				buf += partial;
			} 
		else 
			{
				ret = 0;
				break;
			}
		}
	return ret ? ret : bytes_read;
}

ssize_t f_fx2000_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos)
{
	struct usb_device *udev;
	struct fx2000_device *scanner;
	BOOL minor_presence;
	int minor,i;

	unsigned long copy_size;
	unsigned long bytes_written = 0;
	unsigned int partial;

	int result = 0;
	int maxretry;
	int errn = 0;
	wait_queue_head_t wait_q;

	minor_presence=(BOOL)*((BOOL*)filp->private_data);

	if (!minor_presence)
		return -EFAULT;	// Something wrong: device not more available!

	// Search minor number (for dev_fx2000 index)
	for (i=0; (i<FX2000_NR_DEVS) && (&fx2000_presence[i]!=filp->private_data); i++);
	if (&fx2000_presence[i]!=filp->private_data) return -EFAULT;	// Not Find!!
	minor=i;

	scanner=dev_fx2000[minor];
	udev=scanner->udev;

	init_waitqueue_head(&wait_q);
	
	do 
	{
		unsigned long thistime;
		char *obuf = scanner->bufferOut;

		thistime = copy_size = (count >= BUFFER_OUT_SIZE) ? BUFFER_OUT_SIZE : count;
		if (copy_from_user(obuf, buf, copy_size)) 
		{
			errn = -EFAULT;
			goto error;
		}

		maxretry = 2;
		while (thistime) 
		{
			result = usb_bulk_msg(udev, usb_sndbulkpipe(udev, 6), obuf, thistime, &partial, (int) WR_NAK_TIMEOUT);

#ifdef FX2000_DEBUG
			FX2000_DEBUG("write stat(%i): result:%d thistime:%lu partial:%u\n",minor,result, thistime, partial);
#endif

			if (result == -ETIMEDOUT) /* NAK - so hold for a while */
				{	
					if (!maxretry--) 
					{
						errn = -ETIME;
						goto error;
					}
					//interruptible_sleep_on_timeout(&wait_q, WR_NAK_TIMEOUT);
					continue;
				} 
				else 
					if (!result & partial) 
					{
						obuf += partial;
						thistime -= partial;
					} 
					else
						break;
		};
		if (result) 
		{
#ifdef FX2000_DEBUG
			FX2000_DEBUG("Write Whoops - %x", result);
#endif
			errn = -EIO;
			goto error;
		}
		bytes_written += copy_size;
		count -= copy_size;
		buf += copy_size;
	} while (count > 0);

	return bytes_written ? bytes_written : -EIO;

error:
	return errn;
}

/* Device handled by this driver */
static struct usb_device_id fx2000_id_table [] = {
	{ USB_DEVICE(0x0c43, 0x0200) },
	{ USB_DEVICE(0x0c43, 0x0201) },
	{ USB_DEVICE(0x0c43, 0x0202) },
	{ USB_DEVICE(0x0c43, 0x0203) },
	{ USB_DEVICE(0x0c43, 0x0204) },
	{ USB_DEVICE(0x0c43, 0x0205) },
	{ USB_DEVICE(0x0c43, 0x0206) },
	{ USB_DEVICE(0x0c43, 0x0207) },
	{ USB_DEVICE(0x0c43, 0x0208) },
	{ USB_DEVICE(0x0c43, 0x0209) },
	{ }						/* Terminating entry */
};

#ifdef CONFIG_DEVFS_FS
devfs_handle_t fx2000_devfs_dir;
#endif

//static char devname[4];
MODULE_DEVICE_TABLE (usb,fx2000_id_table);

/*
 * The callbacks are registered within the USB subsystem using the
 * usb_driver data structure
 */
static struct usb_driver fx2000_usb_driver = {
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,4,20)
	.owner=		THIS_MODULE,
#endif
	.name=								"Fx2000",
	.probe=							fx2000_probe,
	.disconnect=					fx2000_disconnect,
	.id_table=						fx2000_id_table,
};



/*
 * Functions called at module load and unload time: only register and
 * unregister the USB callbacks
 */
static int __init fx2000_init(void)
{
	int result, i;

#ifdef CONFIG_DEVFS_FS
	SET_MODULE_OWNER(&fx2000_fops);	
	result = devfs_register_chrdev(fx2000_major, "fx2000", &fx2000_fops);
#else
	result = register_chrdev(fx2000_major, "fx2000", &fx2000_fops);
#endif

#ifdef CONFIG_SYSFS
	sysfs_class = class_simple_create(THIS_MODULE,"fx2000");
#endif
	if (result < 0) {
			FX2000_DEBUG("can't get major %d\n",fx2000_major);
			return result;
	}
	if (fx2000_major == 0) fx2000_major = result; 

#ifdef FX2000_DEBUG
		FX2000_DEBUG("fx2000_major=%i, max device=%i\n",fx2000_major,FX2000_NR_DEVS);
#endif

	
	for (i=0; i < FX2000_NR_DEVS; i++) {
		dev_fx2000[i]=NULL;
		fx2000_presence[i]=FALSE;
	}


#ifndef FX2000_DEBUG
	EXPORT_NO_SYMBOLS;
#endif

	fx2000_create_proc();

	
	return usb_register(&fx2000_usb_driver);

}

static void __exit fx2000_exit(void)
{
	int i;
#ifdef CONFIG_DEVFS_FS	
	devfs_unregister_chrdev(fx2000_major, "fx2000");
#else
	unregister_chrdev(fx2000_major, "fx2000");
#endif
	usb_deregister(&fx2000_usb_driver);

#ifdef CONFIG_SYSFS
	class_simple_destroy(sysfs_class);
#endif

	/* cleanup_module is never called if registering failed */

	fx2000_remove_proc();

	/* Not really necessary */
	for (i=0; i < FX2000_NR_DEVS; i++) {
		if (dev_fx2000[i]!=NULL)
			kfree (dev_fx2000[i]);
	}
}

MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_LICENSE("GPL");

module_init(fx2000_init);
module_exit(fx2000_exit);
